/****************************************************************************
 * Functions for dealing with window borders.
 * Copyright (C) 2004 Joe Wingbermuehle
 ****************************************************************************/

#include "jwm.h"

static char x_bitmap[] = {
	0x00, 0x00, 0x00, 0x00, 0x0C, 0x30, 0x18, 0x18, 0x30, 0x0C, 0x60, 0x06,
	0xC0, 0x03, 0x80, 0x01, 0xC0, 0x03, 0x60, 0x06, 0x30, 0x0C, 0x18, 0x18,
	0x0C, 0x30, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
};

static char min_bitmap[] = {
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFC, 0x3F,
	0xFC, 0x3F, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
};

static char max_bitmap[] = {
	0x00, 0x00, 0x00, 0x00, 0xFC, 0x3F, 0xFC, 0x3F, 0x04, 0x20, 0x04, 0x20,
	0x04, 0x20, 0x04, 0x20, 0x04, 0x20, 0x04, 0x20, 0x04, 0x20, 0x04, 0x20,
	0xFC, 0x3F, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
};

static char maxa_bitmap[] = {
	0x00, 0x00, 0x00, 0x00, 0xFC, 0x3F, 0xFC, 0x3F, 0x04, 0x20, 0xE4, 0x27,
	0xE4, 0x27, 0x24, 0x24, 0x24, 0x24, 0x24, 0x24, 0xE4, 0x27, 0x04, 0x20,
	0xFC, 0x3F, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
};

static void DrawButtonBorder(Drawable canvas, GC gc, ClientNode *np,
	int offset);

/****************************************************************************
 ****************************************************************************/
BorderActionType GetBorderActionType(ClientNode *np, int x, int y) {
	int north;
	int offset;

	if(!(np->borderFlags & BORDER_OUTLINE)) {
		return BA_NONE;
	}

	if(np->borderFlags & BORDER_TITLE) {

		if(y > BORDER_SIZE && y <= BORDER_SIZE + EDGE_SIZE) {
			offset = np->width - EDGE_SIZE + 4;
			if(np->borderFlags & BORDER_CLOSE) {
				if(x > offset && x < offset + EDGE_SIZE) {
					return BA_CLOSE;
				}
				offset -= EDGE_SIZE;
			}
			if(np->borderFlags & BORDER_MAX) {
				if(x > offset && x < offset + EDGE_SIZE) {
					return BA_MAXIMIZE;
				}
				offset -= EDGE_SIZE;
			}
			if(np->borderFlags & BORDER_MIN) {
				if(x > offset && x < offset + EDGE_SIZE) {
					return BA_MINIMIZE;
				}
			}
		}

		if(y >= BORDER_SIZE && y <= BORDER_SIZE + TITLE_SIZE) {
			if(x >= BORDER_SIZE && x < np->width - BORDER_SIZE) {
				if(np->borderFlags & BORDER_MOVE) {
					return BA_MOVE;
				} else {
					return BA_NONE;
				}
			}
		}

		north = TITLE_SIZE;
	} else {
		north = BORDER_SIZE;
	}

	if(!(np->borderFlags & BORDER_RESIZE)) {
		return BA_NONE;
	}

	if(x < EDGE_SIZE && y < EDGE_SIZE) {
		return  BA_RESIZE_N | BA_RESIZE_W | BA_RESIZE;
	} else if(x < EDGE_SIZE && y - north >= np->height - EDGE_SIZE) {
		return BA_RESIZE_S | BA_RESIZE_W | BA_RESIZE;
	} else if(x - BORDER_SIZE >= np->width - EDGE_SIZE
		&& y < EDGE_SIZE) {
		return BA_RESIZE_N | BA_RESIZE_E | BA_RESIZE;
	} else if(x - BORDER_SIZE >= np->width - EDGE_SIZE
		&& y - north >= np->height - EDGE_SIZE) {
		return BA_RESIZE_S | BA_RESIZE_E | BA_RESIZE;
	} else if(x < BORDER_SIZE) {
		return BA_RESIZE_W | BA_RESIZE;
	} else if(x >= np->width) {
		return BA_RESIZE_E | BA_RESIZE;
	} else if(y < BORDER_SIZE) {
		return BA_RESIZE_N | BA_RESIZE;
	} else if(y >= np->height) {
		return BA_RESIZE_S | BA_RESIZE;
	} else {
		return BA_NONE;
	}
}

/****************************************************************************
 ****************************************************************************/
void DrawBorder(ClientNode *np) {
	unsigned int width, height;
	long borderPixel, borderTextPixel;
	long pixelUp, pixelDown;
	int buttonCount, titleWidth;
	XRectangle rect;
	Pixmap canvas;
	GC canvasGC;

	if(shouldExit || !(np->statusFlags & STAT_MAPPED)) {
		return;
	}

	if(!(np->borderFlags & BORDER_TITLE)
		&& !(np->borderFlags & BORDER_OUTLINE)) {
		return;
	}

	width = np->width + BORDER_SIZE + BORDER_SIZE;
	if(np->borderFlags & BORDER_TITLE) {
		height = np->height + TITLE_SIZE + BORDER_SIZE;
	} else {
		height = np->height + 2 * BORDER_SIZE;
	}

	if(np->statusFlags & STAT_ACTIVE) {
		borderPixel = colors[COLOR_ACTIVE_BORDER];
		borderTextPixel = colors[COLOR_ACTIVE_BORDER_TEXT];
		pixelUp = colors[COLOR_ACTIVE_BORDER_UP];
		pixelDown = colors[COLOR_ACTIVE_BORDER_DOWN];
	} else {
		borderPixel = colors[COLOR_BORDER];
		borderTextPixel = colors[COLOR_BORDER_TEXT];
		pixelUp = colors[COLOR_BORDER_UP];
		pixelDown = colors[COLOR_BORDER_DOWN];
	}

	canvas = XCreatePixmap(display, np->parent, width, height,
		DefaultDepth(display, DefaultScreen(display)));
	canvasGC = XCreateGC(display, canvas, 0, NULL);

	XSetForeground(display, canvasGC, borderPixel);
	XFillRectangle(display, canvas, canvasGC, 2, 2,
		width - 4, height - 4);

	buttonCount = DrawBorderButtons(np, canvas, canvasGC);
	titleWidth = width - EDGE_SIZE * buttonCount - BORDER_SIZE * 2 - 2;

	if(np->name && np->name[0] && (np->borderFlags & BORDER_TITLE)) {

		rect.x = BORDER_SIZE + 1;
		rect.y = BORDER_SIZE + 1;
		if(titleWidth > 0) {
			rect.width = titleWidth;
		} else {
			rect.width = 0;
		}
		rect.height = EDGE_SIZE;
		XSetClipRectangles(display, canvasGC, 0, 0, &rect, 1, Unsorted);

		XSetForeground(display, canvasGC, borderTextPixel);
		XSetBackground(display, canvasGC, borderPixel);
		XSetFont(display, canvasGC, borderFont->fid);
		XDrawString(display, canvas, canvasGC,
			BORDER_SIZE + 4,
			borderFont->ascent + borderFont->descent + 3,
			np->name, strlen(np->name));

		XSetClipMask(display, canvasGC, None);

	}

	/* Draw title outline */
	XSetForeground(display, canvasGC, pixelUp);
	XDrawLine(display, canvas, canvasGC, 5, 5, width - 6, 5);
	XDrawLine(display, canvas, canvasGC, 5, 6, 5, EDGE_SIZE + 4);
	XSetForeground(display, canvasGC, pixelDown);
	XDrawLine(display, canvas, canvasGC, 6, EDGE_SIZE + 4, width - 7,
		EDGE_SIZE + 4);
	XDrawLine(display, canvas, canvasGC, width - 6, 6, width - 6,
		EDGE_SIZE + 5);

	/* Draw outline */
	XSetForeground(display, canvasGC, pixelUp);
	XDrawLine(display, canvas, canvasGC, width - 5, 5, width - 5, height - 5);
	XDrawLine(display, canvas, canvasGC, 5, height - 5, width - 5, height - 5);
	XSetForeground(display, canvasGC, pixelDown);
	XDrawLine(display, canvas, canvasGC, 4, 4, width - 5, 4);
	XDrawLine(display, canvas, canvasGC, 4, 5, 4, height - 5);
	XFillRectangle(display, canvas, canvasGC, width - 2, 0, 2, height);
	XFillRectangle(display, canvas, canvasGC, 0, height - 2, width, 2);
	XSetForeground(display, canvasGC, pixelUp);
	XDrawLine(display, canvas, canvasGC, 0, 0, 0, height - 1);
	XDrawLine(display, canvas, canvasGC, 1, 1, 1, height - 2);
	XDrawLine(display, canvas, canvasGC, 1, 0, width - 1, 0);
	XDrawLine(display, canvas, canvasGC, 1, 1, width - 2, 1);

	/* Draw marks */
	XSetForeground(display, canvasGC, pixelDown);

	/* Upper left */
	XDrawLine(display, canvas, canvasGC, EDGE_SIZE + BORDER_SIZE - 1,
		2, EDGE_SIZE + BORDER_SIZE - 1, 3);
	XDrawLine(display, canvas, canvasGC, 2, EDGE_SIZE + BORDER_SIZE - 1,
		3, EDGE_SIZE + BORDER_SIZE - 1);

	/* Upper right */
	XDrawLine(display, canvas, canvasGC, width - EDGE_SIZE - BORDER_SIZE - 1,
		2, width - EDGE_SIZE - BORDER_SIZE - 1, 3);
	XDrawLine(display, canvas, canvasGC, width - 3,
		EDGE_SIZE + BORDER_SIZE - 1, width - 4, EDGE_SIZE + BORDER_SIZE - 1);

	/* Lower left */
	XDrawLine(display, canvas, canvasGC, 2,
		height - EDGE_SIZE - BORDER_SIZE - 1, 3,
		height - EDGE_SIZE - BORDER_SIZE - 1);
	XDrawLine(display, canvas, canvasGC, EDGE_SIZE + BORDER_SIZE - 1,
		height - 3, EDGE_SIZE + BORDER_SIZE - 1, height - 4);

	/* Lower right */
	XDrawLine(display, canvas, canvasGC, width - 3,
		height - EDGE_SIZE - BORDER_SIZE - 1, width - 4,
		height - EDGE_SIZE - BORDER_SIZE - 1);
	XDrawLine(display, canvas, canvasGC, width - EDGE_SIZE - BORDER_SIZE - 1,
		height - 3, width - EDGE_SIZE - BORDER_SIZE - 1, height - 4);

	XSetForeground(display, canvasGC, pixelUp);

	/* Upper left */
	XDrawLine(display, canvas, canvasGC, EDGE_SIZE + BORDER_SIZE,
		2, EDGE_SIZE + BORDER_SIZE, 3);
	XDrawLine(display, canvas, canvasGC, 2, EDGE_SIZE + BORDER_SIZE,
		3, EDGE_SIZE + BORDER_SIZE);

	/* Upper right */
	XDrawLine(display, canvas, canvasGC, width - EDGE_SIZE - BORDER_SIZE, 2,
		width - EDGE_SIZE - BORDER_SIZE, 3);
	XDrawLine(display, canvas, canvasGC, width - 3, EDGE_SIZE + BORDER_SIZE,
		width - 4, EDGE_SIZE + BORDER_SIZE);

	/* Lower left */
	XDrawLine(display, canvas, canvasGC, 2, height - EDGE_SIZE - BORDER_SIZE,
		3, height - EDGE_SIZE - BORDER_SIZE);
	XDrawLine(display, canvas, canvasGC, EDGE_SIZE + BORDER_SIZE,
		height - 3, EDGE_SIZE + BORDER_SIZE, height - 4);

	/* Lower right */
	XDrawLine(display, canvas, canvasGC, width - 3,
		height - EDGE_SIZE - BORDER_SIZE, width - 4,
		height - EDGE_SIZE - BORDER_SIZE);
	XDrawLine(display, canvas, canvasGC, width - EDGE_SIZE - 5,
		height - 3, width - EDGE_SIZE - 5, height - 4);

	XCopyArea(display, canvas, np->parent, np->parentGC, 0, 0,
		width, height, 0, 0);

	XFreeGC(display, canvasGC);
	XFreePixmap(display, canvas);

}

/****************************************************************************
 ****************************************************************************/
void DrawButtonBorder(Drawable canvas, GC gc, ClientNode *np, int offset) {
	long up, down;

	if(np->statusFlags & STAT_ACTIVE) {
		up = colors[COLOR_ACTIVE_BORDER_UP];
		down = colors[COLOR_ACTIVE_BORDER_DOWN];
	} else {
		up = colors[COLOR_BORDER_UP];
		down = colors[COLOR_BORDER_DOWN];
	}

	XSetForeground(display, gc, up);
	XDrawLine(display, canvas, gc, offset, 6, offset, EDGE_SIZE + 3);

	XSetForeground(display, gc, down);
	XDrawLine(display, canvas, gc, offset - 1, 6, offset - 1, EDGE_SIZE + 3);

}

/****************************************************************************
 ****************************************************************************/
int DrawBorderButtons(ClientNode *np, Drawable canvas, GC gc) {
	Pixmap pixmap;
	int count = 0;
	unsigned int offset;
	long fg, bg;

	if(!(np->borderFlags & BORDER_TITLE)) {
		return count;
	}

	offset = np->width + BORDER_SIZE - EDGE_SIZE;

	if(np->statusFlags & STAT_ACTIVE) {
		fg = colors[COLOR_ACTIVE_BORDER_TEXT];
		bg = colors[COLOR_ACTIVE_BORDER];
	} else {
		fg = colors[COLOR_BORDER_TEXT];
		bg = colors[COLOR_BORDER];
	}

	if(np->borderFlags & BORDER_CLOSE) {
		DrawButtonBorder(canvas, gc, np, offset);

		pixmap = XCreatePixmapFromBitmapData(display, canvas, x_bitmap,
			16, 16, fg, bg, DefaultDepth(display, DefaultScreen(display)));
		XCopyArea(display, pixmap, canvas, gc, 0, 0, 16, 16,
			offset + 2, 8);
		XFreePixmap(display, pixmap);

		offset -= EDGE_SIZE;
		++count;
	}

	if(np->borderFlags & BORDER_MAX) {

		if(np->statusFlags & STAT_MAXIMIZED) {
			pixmap = XCreatePixmapFromBitmapData(display, canvas, maxa_bitmap,
				16, 16, fg, bg, DefaultDepth(display, DefaultScreen(display)));
		} else {
			pixmap = XCreatePixmapFromBitmapData(display, canvas, max_bitmap,
				16, 16, fg, bg, DefaultDepth(display, DefaultScreen(display)));
		}
		XCopyArea(display, pixmap, canvas, gc, 0, 0, 16, 16,
			offset + 2, 8);
		XFreePixmap(display, pixmap);

		DrawButtonBorder(canvas, gc, np, offset);

		offset -= EDGE_SIZE;
		++count;
	}

	if(np->borderFlags & BORDER_MIN) {
		DrawButtonBorder(canvas, gc, np, offset);
		pixmap = XCreatePixmapFromBitmapData(display, canvas, min_bitmap,
			16, 16, fg, bg, DefaultDepth(display, DefaultScreen(display)));
		XCopyArea(display, pixmap, canvas, gc, 0, 0, 16, 16,
			offset + 2, 8);
		XFreePixmap(display, pixmap);
		++count;
	}

	return count;

}



