/***************************************************************************
 * Functions to handle the tray.
 * Copyright (C) 2004 Joe Wingbermuehle
 ***************************************************************************/

#include "jwm.h"

typedef struct TrayNode {
	ClientNode *client;
	struct TrayNode *next;
} TrayNode;

const int TRAY_HEIGHT = 28;
static const int TRAY_ITEM_HEIGHT = 18;
static const int TRAY_START = 64;
static const int TRAY_TEXT_START = 26;

static char min_bitmap[] = {
	0xFF, 0xFF, 0x01, 0x80, 0x0D, 0xB0, 0x1D, 0xB8, 0x39, 0x9C, 0x71, 0x8E,
	0xE1, 0x87, 0xC1, 0x83, 0xC1, 0x83, 0xE1, 0x87, 0x71, 0x8E, 0x39, 0x9C,
	0x1D, 0xB8, 0x0D, 0xB0, 0x01, 0x80, 0xFF, 0xFF
};

static char max_bitmap[] = {
	0xFF, 0xFF, 0x01, 0x80, 0x01, 0x80, 0x01, 0x80, 0x01, 0x80, 0x01, 0x80,
	0x01, 0x80, 0x01, 0x80, 0x01, 0x80, 0x01, 0x80, 0x01, 0x80, 0x01, 0x80,
	0x01, 0x80, 0x01, 0x80, 0x01, 0x80, 0xFF, 0xFF
};

static char *menuTitle = "JWM";

Window trayWindow;
static GC trayGC;
static Pixmap buffer; 
static GC bufferGC;
static unsigned int trayStop;
static unsigned int trayTextOffset;

static TrayNode *trayNodes;

static unsigned int GetItemCount();
static unsigned int GetItemWidth();
static int ShouldShowItemInTray(ClientNode *np);

/***************************************************************************
 ***************************************************************************/
void CreateTray() {
	XSetWindowAttributes attr;

	attr.override_redirect = True;
	attr.event_mask = ButtonPressMask | ExposureMask | KeyPressMask;
	attr.background_pixel = colors[COLOR_BORDER];

	trayWindow = XCreateWindow(display, rootWindow, 0,
		rootHeight - TRAY_HEIGHT - 1, rootWidth - 1, TRAY_HEIGHT, 1,
		DefaultDepth(display, DefaultScreen(display)), InputOutput,
		DefaultVisual(display, DefaultScreen(display)),
		CWOverrideRedirect | CWBackPixel | CWEventMask, &attr);
	XMapRaised(display, trayWindow);

	SetDefaultCursor(trayWindow);

	trayGC = XCreateGC(display, trayWindow, 0, NULL);
	XSetFont(display, trayGC, trayFont->fid);

	buffer = XCreatePixmap(display, rootWindow,
		rootWidth - trayStop - 1, TRAY_HEIGHT,
		DefaultDepth(display, DefaultScreen(display)));
	bufferGC = XCreateGC(display, buffer, 0, 0);
	XSetFont(display, bufferGC, trayFont->fid);

	trayStop = XTextWidth(trayFont, " 00:00 MM", 9) + LOAD_WIDTH + 5;

	trayTextOffset = (TRAY_HEIGHT / 2) + (trayFont->ascent / 2);

	trayNodes = NULL;

}

/***************************************************************************
 ***************************************************************************/
void AddClientToTray(ClientNode *np) {
	TrayNode *tp;

	tp = Allocate(sizeof(TrayNode));
	tp->client = np;
	tp->next = trayNodes;
	trayNodes = tp;

	DrawTray();
}

/***************************************************************************
 ***************************************************************************/
void RemoveClientFromTray(ClientNode *np) {
	TrayNode *tp, *lp;

	lp = NULL;
	for(tp = trayNodes; tp; tp = tp->next) {
		if(tp->client == np) {
			if(lp) {
				lp->next = tp->next;
			} else {
				trayNodes = tp->next;
			}
			Release(tp);
			break;
		}
		lp = tp;
	}

	DrawTray();
}

/***************************************************************************
 ***************************************************************************/
void FocusNext() {
	TrayNode *tp;

	for(tp = trayNodes; tp; tp = tp->next) {
		if(!(tp->client->statusFlags & STAT_STICKY)
			&& currentDesktop != tp->client->desktop) {
			continue;
		}
		if(tp->client->statusFlags & STAT_ACTIVE) {
			if(tp->next) {
				tp = tp->next;
			} else {
				tp = trayNodes;
			}
			break;
		}
	}
	if(!tp) {
		tp = trayNodes;
	}

	while(tp && (tp->client->desktop != currentDesktop)
		&& !(tp->client->statusFlags & STAT_STICKY)) {
		tp = tp->next;
	}
	if(!tp) {
		tp = trayNodes;
		while(tp && (tp->client->desktop != currentDesktop)
			&& !(tp->client->statusFlags & STAT_STICKY)) {
			tp = tp->next;
		}
	}

	if(tp) {
		RestoreClient(tp->client);
		FocusClient(tp->client);
	}

}

/***************************************************************************
 ***************************************************************************/
void DestroyTray() {

	XFreeGC(display, trayGC);
	XFreeGC(display, bufferGC);
	XFreePixmap(display, buffer);

	XDestroyWindow(display, trayWindow);
}

/***************************************************************************
 ***************************************************************************/
void HandleTrayClick(XButtonEvent *event) {
	TrayNode *tp;
	int x, y;
	int realx, realy;
	int index;
	unsigned int width;

	XRaiseWindow(display, trayWindow);
	XAllowEvents(display, ReplayPointer, CurrentTime);

	GetMousePosition(&realx, &realy);

	y = realy - rootHeight + TRAY_HEIGHT + 1;
	x = realx;

	width = GetItemWidth(GetItemCount());

	if(y > 1) {

		if(x < TRAY_START) {
			ShowRootMenu(0, rootHeight - TRAY_HEIGHT);
			return;
		} else {
			x -= TRAY_START;
		}

		if(x < pagerWidth) {
			ChangeDesktop(x / (pagerWidth / desktopCount));
			return;
		} else {
			x -= pagerWidth + 1;
		}

		index = x / (int)width;
		for(tp = trayNodes; tp; tp = tp->next) {
			if(ShouldShowItemInTray(tp->client)) {
				if(index == 0) {
					switch(event->button) {
					case Button1:
						RestoreClient(tp->client);
						FocusClient(tp->client);
						break;
					case Button3:
						ShowWindowMenu(tp->client, realx, realy);
						break;
					default:
						break;
					}
					DrawTray();
					break;
				}
				--index;
			}
		}
	}
}

/***************************************************************************
 ***************************************************************************/
void ChangeDesktop(int desktop) {
	ClientNode *np;
	long lvalue;

	if(desktop >= desktopCount || desktop < 0) {
		return;
	}

	if(currentDesktop == desktop) {
		return;
	}

	for(np = nodes; np; np = np->next) {
		if(np->statusFlags & STAT_STICKY) {
			continue;
		}
		if(np->desktop == desktop) {
			ShowClient(np);
		} else if(np->desktop == currentDesktop) {
			HideClient(np);
		}
	}

	currentDesktop = desktop;

	lvalue = currentDesktop;
	XChangeProperty(display, rootWindow, protNetCurrentDesktop,
		XA_CARDINAL, 32, PropModeReplace,
		(unsigned char*)&lvalue, 1);

	lvalue = currentDesktop;
	XChangeProperty(display, rootWindow, protWinWorkspace,
		XA_CARDINAL, 32, PropModeReplace,
		(unsigned char*)&lvalue, 1);

	DrawTray();
}

/***************************************************************************
 ***************************************************************************/
void DrawTray() {
	Pixmap pixmap;
	TrayNode *tp;
	unsigned int width, x, y;
	unsigned int remainder;
	long borderPixel, borderTextPixel;
	unsigned int itemCount;

	if(shouldExit) {
		return;
	}

	XSetForeground(display, bufferGC, colors[COLOR_BORDER]);
	XFillRectangle(display, buffer, bufferGC, 0, 0,
		rootWidth - trayStop - 1, TRAY_HEIGHT);

	XSetForeground(display, bufferGC, colors[COLOR_BORDER_DOWN]);
	XDrawRectangle(display, buffer, bufferGC, 1, 1,
		TRAY_START - 3, TRAY_HEIGHT - 3);
	XSetForeground(display, bufferGC, colors[COLOR_BORDER_UP]);
	XDrawRectangle(display, buffer, bufferGC, 2, 2,
		TRAY_START - 3, TRAY_HEIGHT - 3);

	XSetForeground(display, bufferGC, colors[COLOR_BORDER_DOWN]);
	x = (TRAY_START - 3) / 2
		- XTextWidth(trayFont, menuTitle, strlen(menuTitle)) / 2;
	y = (TRAY_HEIGHT - 3) / 2 + (trayFont->ascent + trayFont->descent) / 2;
	XDrawString(display, buffer, bufferGC, x, y,
		menuTitle, strlen(menuTitle));
	XSetForeground(display, bufferGC, colors[COLOR_BORDER_TEXT]);
	XDrawString(display, buffer, bufferGC, x - 1, y - 1,
		menuTitle, strlen(menuTitle));

	itemCount = GetItemCount();

	if(!itemCount) {
		UpdatePager();
		UpdateTime();
		return;
	}

	width = GetItemWidth(itemCount);

	x = TRAY_START + pagerWidth + 1;
	for(tp = trayNodes; tp; tp = tp->next) {

		if(!ShouldShowItemInTray(tp->client)) {
			continue;
		}

		if(tp->client->statusFlags & STAT_ACTIVE) {
			borderPixel = colors[COLOR_ACTIVE_BORDER];
			borderTextPixel = colors[COLOR_ACTIVE_BORDER_TEXT];
		} else {
			borderPixel = colors[COLOR_BORDER];
			borderTextPixel = colors[COLOR_BORDER_TEXT];
		}

		XSetForeground(display, bufferGC, borderPixel);
		XFillRectangle(display, buffer, bufferGC, x, 2,
			width - 2, TRAY_HEIGHT - 4);

		if(!(tp->client->statusFlags & STAT_MINIMIZED)) {
			pixmap = XCreatePixmapFromBitmapData(display, buffer, max_bitmap,
				16, 16, colors[COLOR_BORDER_TEXT],
				colors[COLOR_BORDER], DefaultDepth(display,
				DefaultScreen(display)));
		} else {
			pixmap = XCreatePixmapFromBitmapData(display, buffer, min_bitmap,
				16, 16, colors[COLOR_BORDER_TEXT],
				colors[COLOR_BORDER],
				DefaultDepth(display, DefaultScreen(display)));
		}

		XCopyArea(display, pixmap, buffer, bufferGC, 0, 0, 16, 16,
			x + 5, 6);
		XFreePixmap(display, pixmap);

		if(tp->client->name && tp->client->name[0]) {
			XSetForeground(display, bufferGC, borderTextPixel);
			XSetBackground(display, bufferGC, borderPixel);
			XDrawString(display, buffer, bufferGC,
				x + TRAY_TEXT_START, trayTextOffset,
				tp->client->name, strlen(tp->client->name));
		}
		XSetForeground(display, bufferGC, colors[COLOR_BORDER]);
		XFillRectangle(display, buffer, bufferGC,
			x + width - 1, 3, 2, TRAY_HEIGHT - 5);

		XSetForeground(display, bufferGC, colors[COLOR_BORDER_DOWN]);
		XDrawRectangle(display, buffer, bufferGC, x, 1, width - 3,
			TRAY_HEIGHT - 4);

		XSetForeground(display, bufferGC, colors[COLOR_BORDER_UP]);
		XDrawRectangle(display, buffer, bufferGC, x + 1, 2,
			width - 3, TRAY_HEIGHT - 4);

		x += width;
	}

	remainder = rootWidth - TRAY_START - pagerWidth - 1
		- trayStop - (width * itemCount);
	if(remainder) {
		XSetForeground(display, bufferGC, colors[COLOR_BORDER]);
		XFillRectangle(display, buffer, bufferGC,
			x, 3, remainder, TRAY_HEIGHT - 5);
	}

	UpdatePager();
	UpdateTime();

}

/***************************************************************************
 ***************************************************************************/
void UpdatePager() {
	XCopyArea(display, GetPager(), buffer, bufferGC,
		0, 0, pagerWidth, pagerHeight, TRAY_START, 1);
	XCopyArea(display, buffer, trayWindow, trayGC,
		0, 0, rootWidth - trayStop - 1, TRAY_HEIGHT - 1, 0, 0);
}

/***************************************************************************
 ***************************************************************************/
void UpdateTime() {
	static char tstring[80];
#ifdef SHOW_LOAD
	Pixmap loadDisplay;
#endif
	unsigned int len;
	time_t clock;
	struct tm *timeInfo;
	unsigned int x;

	clock = time(NULL);
	timeInfo = localtime(&clock);
	len = strftime(tstring, sizeof(tstring), "%I:%M %p", timeInfo);
	if(tstring[0] == '0') {
		for(x = 0; x < len; x++) {
			tstring[x] = tstring[x + 1];
		}
		--len;
	}
	x = rootWidth - XTextWidth(trayFont, tstring, len) - 4;

	XSetForeground(display, trayGC, colors[COLOR_BORDER]);
	XFillRectangle(display, trayWindow, trayGC,
		rootWidth - trayStop, 0,
		trayStop, TRAY_HEIGHT);

	XSetBackground(display, trayGC, colors[COLOR_BORDER]);
	XSetForeground(display, trayGC, colors[COLOR_BORDER_TEXT]);
	XDrawString(display, trayWindow, trayGC, x,
		trayTextOffset, tstring, len);

#ifdef SHOW_LOAD
	XSetForeground(display, trayGC, colors[COLOR_BORDER_DOWN]);
	XDrawRectangle(display, trayWindow, trayGC, rootWidth - trayStop,
		1, LOAD_WIDTH + 1, TRAY_HEIGHT - 3);

	loadDisplay = GetLoadDisplay();
	XCopyArea(display, loadDisplay, trayWindow, trayGC, 0, 0,
		LOAD_WIDTH, TRAY_HEIGHT - 4, rootWidth - trayStop + 1, 2);
#endif

}

/***************************************************************************
 ***************************************************************************/
unsigned int GetItemCount() {
	TrayNode *tp;
	unsigned int count;

	count = 0;
	for(tp = trayNodes; tp; tp = tp->next) {
		if(ShouldShowItemInTray(tp->client)) {
			++count;
		}
	}

	return count;
}

/***************************************************************************
 ***************************************************************************/
int ShouldShowItemInTray(ClientNode *np) {

	if(np->desktop != currentDesktop
		&& !(np->statusFlags & STAT_STICKY)) {
		return 0;
	}

	if(np->owner != None || (np->statusFlags & STAT_NOLIST)
		|| (np->statusFlags & STAT_WITHDRAWN)
		|| (!(np->statusFlags & STAT_MAPPED)
		&& !(np->statusFlags & STAT_MINIMIZED))) {
		return 0;
	}

	return 1;
}

/***************************************************************************
 ***************************************************************************/
unsigned int GetItemWidth(unsigned int itemCount) {
	unsigned int width;

	width = rootWidth - TRAY_START - pagerWidth - 1 - trayStop;

	if(!itemCount) {
		return width;
	}

	width /= itemCount;
	if(!width) {
		width = 1;
	}

	return width;
}

